
from http.server import BaseHTTPRequestHandler, HTTPServer
import os

UPLOAD_FOLDER = 'uploads'
if not os.path.exists(UPLOAD_FOLDER):
    os.makedirs(UPLOAD_FOLDER)

class FileServer(BaseHTTPRequestHandler):
    def _set_headers(self, status_code=200, content_type='text/plain'):
        self.send_response(status_code)
        self.send_header('Content-type', content_type)
        self.end_headers()

    def do_POST(self):
        content_length = int(self.headers['Content-Length'])
        filename = self.headers['File-Name']
        file_path = os.path.join(UPLOAD_FOLDER, filename)
        with open(file_path, 'wb') as f:
            f.write(self.rfile.read(content_length))
        self._set_headers(status_code=201)
        self.wfile.write(b'File uploaded successfully')

    def do_GET(self):
        filename = self.path[1:]
        file_path = os.path.join(UPLOAD_FOLDER, filename)
        if os.path.exists(file_path):
            self._set_headers(content_type='application/octet-stream')
            with open(file_path, 'rb') as f:
                self.wfile.write(f.read())
        else:
            self._set_headers(status_code=404)
            self.wfile.write(b'File not found')

def run(server_class=HTTPServer, handler_class=FileServer, port=8000):
    server_address = ('', port)
    httpd = server_class(server_address, handler_class)
    print(f'Starting file server on port {port}...')
    httpd.serve_forever()

if __name__ == '__main__':
    run()


from .Node import nodes
from .Node.diffs import changed_pixel_return_dict as cprd
from .utils.Pixels import image_to_array as itoa
from .utils.Pixels import is_image_changed as iic
from .utils.Pixels import modify_image
from .utils.helper import save_graph , load_graph, show_image
import os
import numpy as np
import click
from .server import pull_file, push_file


def ensure_graphs_directory():
    graphs_dir = os.path.join(os.path.dirname(__file__), 'Graphs')

    # Check if the directory exists
    if not os.path.exists(graphs_dir):
        # Create the directory if it doesn't exist
        os.makedirs(graphs_dir)
        # print(f"Created Graphs directory at: {graphs_dir}")
        return graphs_dir+'/'
    else:
        # print(f"Graphs directory already exists at: {graphs_dir}")
        return graphs_dir+'/'
    


def check_graph_initialise(name):
    '''
    To check if the image already has a graph.pkl file
    cons:
        - image name should not be changed
    It searches name.pkl file in Graphs directory

    returns a bool value
    '''

    _path=ensure_graphs_directory()+name+".pkl"
    check_file_existence=os.path.exists(_path)

    return check_file_existence


def create_root_node(image_path:str,
                     author:str):
    '''
    Function to add a root node and initalise the graph
    Takes image from image_path and convert into numpy.ndarray
    Initalises graphImage with graph_name as image name

    Also, saves the graph at IVCS/Graphs/ using helper.save_graph
    Checks if the same graph already exists

    inputs->[image_path,
             author]
    
    '''

    _graph_name=(image_path.split('/')[-1]).split('.')[0]
    _image=itoa(image_path)

    if(not check_graph_initialise(_graph_name)):
        gr=nodes.imageGraph(image=_image,author=author,graph_name=_graph_name)
        save_graph(gr)
        print(f"The imageGraph has been initialised for {image_path} at {ensure_graphs_directory()+_graph_name}")


    else:
        raise Exception(f"The imageGraph for this image has already been initialised.")



def add_new_node(imageGraph_instance:nodes.imageGraph,
                commit_message:str,
                author:str,
                image:np.ndarray):
    '''
    inputs->[imageGraph object, 
             commit message,
             author,
             image]

    Function to add a new node on the current head location in the nodes.imageGraph object
    Expects an established graph which has a nodes.graphNode object
    if the graph of the image has not been initalised, use create_root_node to initalise
    stores changes in the image in dict format

    returns a nodes.graphImage object, 
    doesn't save the new graph
    '''

    _image_head=imageGraph_instance.return_np_image_at_head()
    _head=imageGraph_instance.Head
    # _image=imageGraph_instance.return_np_image_at_head()
    if(iic(_image_head, image)):
        _change_dict=cprd(_image_head,image)

        new_node=nodes.graphNode(
                                change=_change_dict,
                                author=author,
                                in_nodes=_head,
                                commit_message=commit_message)
        
        _head.out_nodes.append(new_node)
        imageGraph_instance.Head=new_node
        imageGraph_instance.log.append(new_node)
        return imageGraph_instance
    else:
        raise Exception(f"Image has not been changed with respect to the current location of the HEAD in the graphImage")
        


def revert_head_by_one_return_imageGraph(imageGraph_instance:nodes.imageGraph):
    current_head=imageGraph_instance.Head
    if(current_head.in_nodes is None):
        raise Exception(f'The Head is already pointing at the root node there are no previous versions.')
    else:
        reverted_head=current_head.in_nodes

        imageGraph_instance.Head=reverted_head

        print(f"The head has been reverted to \n unique_id: {reverted_head.unique} \n message: {reverted_head.commit_message} \n author :{reverted_head.author}\n time: {reverted_head.time}")
        
        return imageGraph_instance


def revert_by_one_save_imageGraph(image_path:str):
    _imageGraph_path=image_path.split('/')[-1].split('.')[0]
    _imageGraph_instance=load_graph(_imageGraph_path)
    _imageGraph_instance=revert_head_by_one_return_imageGraph(_imageGraph_instance)
    save_graph(_imageGraph_instance)

    d__imageGraph_instance=load_graph(_imageGraph_path)
    _img_to_be_saved=d__imageGraph_instance.save_image()
    modify_image(image_path=image_path, image=_img_to_be_saved)
    



def add_version(image_path:str,
                author:str,
                commit_message:str):
    _image=itoa(image_path)
    _graph_name=((image_path.split('/')[-1]).split('.'))[0]
    _imageGraph_instance=load_graph(_graph_name)
    _imageGraph_instance=add_new_node(imageGraph_instance=_imageGraph_instance,
                                      image=_image,author=author,
                                      commit_message=commit_message)
    save_graph(_imageGraph_instance)

def forward_head_by_one_return_imageGraph(imageGraph_instance:nodes.imageGraph):
    _head=imageGraph_instance.Head
    if(len(_head.out_nodes) == 1):
        imageGraph_instance.Head=_head.out_nodes[0]
        
        print(f"\n The head has been forwarded to \n unique_id: {imageGraph_instance.Head.unique} \n message: {imageGraph_instance.Head.commit_message} \n author :{imageGraph_instance.Head.author} \n time :{imageGraph_instance.Head.time}")
        return imageGraph_instance
    elif(len(_head.out_nodes)>1):
        print(f"\nThere are multiple nodes going out from the current HEAD location \n")
        print(f"The outgoing nodes are : \n")

        for i in range(len(_head.out_nodes)):
            print(i+1, ' Node \n' )
            _print_details(_head.out_nodes[i])

        print(f" To shift the head on any of the graphNode, Use shift command with the respective unique id")
        return False
    else:
        raise Exception(f"There are no outgoing nodes from current HEAD")
    
def forward_head_by_one_save_imageGraph(image_path:str):
    _imageGraph_path=image_path.split('/')[-1].split('.')[0]
    _imageGraph_instance=load_graph(_imageGraph_path)

    if(len(_imageGraph_instance.Head.out_nodes)==1):

        _imageGraph_instance=forward_head_by_one_return_imageGraph(_imageGraph_instance)
        save_graph(_imageGraph_instance)

        d__imageGraph_instance=load_graph(_imageGraph_path) 
        _img_to_be_saved=d__imageGraph_instance.save_image()
        modify_image(image_path=image_path, image=_img_to_be_saved)

    else:
        forward_head_by_one_return_imageGraph(_imageGraph_instance)

    

def show_image_(_graph_name:str):
    _imageGraph=load_graph(_graph_name)
    _imageGraph.show_image()

def _print_details(_head):
    print(f" \n {type(_head)}\n unique id : {_head.unique} \n Commit Message : {_head.commit_message}\n Author : {_head.author} \n Time : {_head.time} \n ")
    


@click.group()
def cli():
    pass


@cli.command()
def init():
    '''
    Checks if the Graphs directory exists to store the pixv logs
    '''
    ensure_graphs_directory()

@cli.command()
@click.argument("image_path")
def hist(image_path):
    '''
    Prints the tree structure of the versions
    '''
    _graph_name=image_path.split('/')[-1].split('.')[0]
    _gr=load_graph(_graph_name)
    print("\n")
    _gr.print_graph_tree(node=_gr.root_node)
    print("\n")

@cli.command()
@click.argument("graph_name")
def locate(graph_name):
    '''
    Specifies the HEAD position
    '''
    print(f"\n The Head is currrently pointing to the following graphNode :\n")
    gr_=load_graph(graph_name)
    _head=gr_.Head
    _print_details(_head)

@cli.command()
@click.argument("image_path")
def logs(image_path):
    '''
    prints the version history for the image
    '''
    _graph_name=image_path.split('/')[-1].split('.')[0]
    _gr=load_graph(_graph_name)
    for i in range(len(_gr.log)):
        _print_details(_gr.log[i])

# @click.argument('IMAGE_PATH')
@cli.command()
@click.argument("graph_name")
def show(graph_name):
    '''
    Loads the image at the current head location doesnt modify the image at the image path
    '''
    print(f"The image is loading")
    show_image_(graph_name)


@cli.command()
# @click.option("--paths", help="Path of the Image")
@click.argument('PATHS')
@click.option("--author", help="Author Name")
def add(paths,author):
    '''
    Intialize the graphImage (creates the rootNode)
    HEAD will point to the rootNode after intialising. 
    '''
    if paths is None or not os.path.isfile(paths):
       click.echo("Error: Invalid image path.")
       return
    else:
        create_root_node(image_path=paths,
                        author=author)
        

@cli.command()
@click.argument('IMAGE_PATH')
@click.option('--author',help="Author Name")
@click.option('--message',help="Commit Message")
def commit(image_path,author,message):
    '''
    Adds new versions to the existing graphImage and shift the HEAD to the new version
    '''

    if image_path is None or not os.path.isfile(image_path):
       click.echo("Error: Invalid image path.")
       return
    else:
        add_version(image_path=image_path,commit_message=message,author=author)
        print("Added a new version.")
        

@cli.command()
@click.argument('IMAGE_PATH')
def revert(image_path):
    '''
    reverts the HEAD by one position and modifies the image at image path according to the HEAD
    '''
    
    if image_path is None or not os.path.isfile(image_path):
       click.echo("Error: Invalid image path.")
       return
    else:
        revert_by_one_save_imageGraph(image_path=image_path)


        print("The head has been reverted by one position")


@cli.command()
@click.argument('IMAGE_PATH')
def forward(image_path):
    '''
    move the HEAD forward by one position and modifies the image at image path according to the HEAD.
    If multiple nodes are going out from the current node, "shift" is to be used.
    '''
    
    if image_path is None or not os.path.isfile(image_path):
       click.echo("Error: Invalid image path.")
       return
    else:
        forward_head_by_one_save_imageGraph (image_path=image_path)


@cli.command()
@click.argument('IMAGE_PATH')
def original(image_path):
    '''
    doesnt shift the head, just shows the original image stored in the rootNode of the graphImage
    '''
    
    if image_path is None or not os.path.isfile(image_path):
       click.echo("Error: Invalid image path.")
       return
    else:
        _name_graph=image_path.split('/')[-1].split('.')[0]
        _gr=load_graph(_name_graph)
        show_image(_gr)


@cli.command()
@click.argument('IMAGE_PATH')
@click.option('--id', help="unique id of the node you want to locate")
def shift(image_path,id):
    '''
    shifts head location according to the id and modifies the image at image path according to HEAD
    '''
    
    if image_path is None or not os.path.isfile(image_path):
       click.echo("Error: Invalid image path.")
       return
    else:
        _name_graph=image_path.split('/')[-1].split('.')[0]
        _gr=load_graph(_name_graph)
        
        _gr.shift_head_by_unique(id)
        save_graph(_gr)

        d__gr=load_graph(_name_graph)
        _img_to_be_saved=d__gr.save_image()
        modify_image(image_path=image_path,image=_img_to_be_saved)
        


@cli.command()
@click.argument("image_path")
def push(image_path):
    _graph_name=image_path.split('/')[-1].split('.')[0]
    _path=ensure_graphs_directory()+_graph_name+'.pkl'
    push_file.upload_file(filename=_path,server_address="127.0.0.1:8000")
    

@cli.command()
@click.argument("image_path")
def pull(image_path):
    _graph_name=image_path.split('/')[-1]
    _path=ensure_graphs_directory()
    pull_file.download_file(filename=_graph_name,server_address="127.0.0.1:8000",download_dir=_path)

if __name__ == "__main__":
    cli()



from PIL import Image
import numpy as np

def image_to_array(image_path):
    # Open the image using Pillow
    image = Image.open(image_path)

    # Convert the image to a NumPy array
    image_array = np.array(image)
    image_array = image_array.astype(np.int64)

    return image_array



def is_image_changed(image1,image2):
    
    return not np.array_equal(image1, image2)

def modify_image(image_path:str,image:np.ndarray):
    try:
        img = Image.fromarray(image)
        img.save(image_path)
        return True
    
    except Exception as e:
        print(f"An error occured :{e}")
        return False




from PIL import Image
import numpy as np


def image_dimensions_return_list(image1:np.ndarray):
    '''
    Function to return the shape of the image in tuple
    '''
    return((np.shape(image1)))


def check_image_dimensions_return_bool(image1:np.ndarray,
                                       image2:np.ndarray):
    '''
    Function to check if the dimensions of the image are still same.
    Inputs->[image1:np.ndarray,image2:np.ndarray]
    ouput->bool
    '''

    return (image1.shape==image2.shape)

# print(check_image_dimensions_return_bool(image1,image2))



# zhape=(image_dimensions_return_list(image1))
# print(zhape)
# _i,_j,_k=zhape
# print(_i)

def changed_pixel_return_dict(head_image:np.ndarray,
                              new_image:np.ndarray):

    '''
    INPUT:-
    image1 and image2 both of np.ndarray dtype

    Function iterates over all the pixels and stores the pixels that have changed in dict

    OUTPUT:-
    the output of this function is a dictionary
    the key comprises of a list that specifies the location of the pixel that changed, and
    the value comprises of the decimal(float) value signifying the change in the pixel 

    '''


 

    if(check_image_dimensions_return_bool(image1=head_image,image2=new_image)):
        _shape=image_dimensions_return_list(head_image)
        changes_dict={}
        changes_count=0

        _i,_j,_k=_shape
        for i in range(_i):
            for j in range(_j):
                for k in range(_k):
                    # count2=count2+1
                    if(head_image[i][j][k]!=new_image[i][j][k]):
                        changes_count=changes_count+1
                        _change=(new_image[i][j][k])-head_image[i][j][k]
                        _list=(i,j,k)

                        changes_dict[_list]=_change
                    else:
                        continue
        return changes_dict
        # print(count2)
        # return count
    else:
        raise Exception(f"The dimensions of the image are not same. PixV isnt compatible with cropping and changing dimensions of image.")
    
# ert=changed_pixel_return_dict(image2,image4)
# print(ert)






import os
from datetime import datetime
import uuid
from ..utils import Pixels
import numpy as np
from PIL import Image
from matplotlib import pyplot as plt

class graphNode:
    def __init__(self,
                 change:dict,
                 in_nodes,
                 author:str,
                 commit_message=None,
                 version_name=None
                 ) :
        self.change=change
        self.version_name=version_name
        self.out_nodes=list()
        self.in_nodes=in_nodes
        self.time=datetime.now()
        self.commit_message=commit_message
        self.author=author
        self.unique=uuid.uuid4()
    
    def dictNode(self):
        node_dict = {
            "Change": self.change,
            "version_name": self.version_name,
            "out_nodes": self.out_nodes,
            "in_nodes": self.in_nodes,
            "time": str(self.time),  # Convert datetime to string
            "commit_message": self.commit_message,
            "author": self.author,
            "unique": str(self.unique)  # Convert UUID to string
        }

        print(node_dict)
    
    
class rootNode(graphNode):
    def __init__(self,
                 image:np.ndarray,
                 author:str,
                 version_name:str,
                 in_nodes=None,
                 commit_message=None,
                 ):
        self.image=image
        self.out_nodes=list()
        super().__init__(change=None,version_name=version_name,author=author,in_nodes=in_nodes,commit_message=commit_message)
        super().dictNode()




class imageGraph:
    def __init__(self,
                 image:np.ndarray,
                 author:str,
                 graph_name:str):
        self.root_node=rootNode(image=image,author=author,version_name=graph_name)
        self.Head=self.root_node #head kaha point kar raha hai
        self.graph_name=graph_name
        self.image=image
        self.log=list()
        self.log.append(self.root_node)

    def find_node_by_unique(self, unique_id, node=None):
        """
        Find a graphNode with a given unique identifier using depth-first search (DFS).

        Args:
        unique_id (str): The unique identifier of the node to find.
        node (graphNode): The starting node for the search. Default is None (start from the root).

        Returns:
        graphNode or None: The node with the given unique identifier if found, else None.
        """
        unique_id=unique_id.strip()
        if node is None:
            node = self.root_node #will start searching from rootNode
            print(f"\nStarting search from rootNode")

        if str(node.unique) == unique_id:
            print(f"graphNode found")
            return node


        if len(node.out_nodes)!=0:
            for child_node in node.out_nodes:
                result = self.find_node_by_unique(unique_id, child_node)
                if result:
                    return result

        return None
    
    def shift_head_by_unique(self, unique_id):
        """
        Shift the head location to the node with the given unique identifier if it exists.

        Args:
        unique_id (str): The unique identifier of the node to shift the head to.

        Returns:
        bool: True if the head location is shifted, False otherwise.
        """
        node = self.find_node_by_unique(unique_id)
        if node:
            self.Head = node
            print(f"Shifted HEAD to {unique_id} \n")
            return True
        else:
            print("Not a valid unique ID.")
            return False

    
    def _traverse_graph(self):

        '''
        Private function to climb up the graph from the current head location to rootNode.
        And calculate the total changes and return a dictionary for the changes.

        Isnt compatible when number of pixels in the nodes have been changed with respect to each other.
        '''
        if(type(self.Head) is not rootNode):
            _head=self.Head
            _all_changes={}
            _all_changes.update(_head.change)

            while(_head.in_nodes is not None):
                _change=_head.in_nodes.change
                if(_change is not None):

                    for key in _change.keys():
                        if(key in _all_changes.keys()):
                            _all_changes[key]+=_change[key]
                        else:
                            _all_changes[key]=_change[key]
                            
                    _head=_head.in_nodes
                else:
                    break



        else:
            _all_changes={}
        return _all_changes
    

    def return_np_image_at_head(self):
        '''
        Function to return complete image in np.ndarray dtype at Head of imageGraph
        this function is intended to be used to calculate the changes in the new version during a commit.
        return image in np.ndarray format.
        
        '''

        _all_changes=self._traverse_graph()
        image_=self.image.copy()


        for key in _all_changes.keys():
            if len(key) == 3:
                i,j,k=key

                image_[i][j][k]+=_all_changes[key]  
            else:
                print("bhaii kya, 4 dimensions kaise image ke?")

        return image_ #np.ndarray


    def print_graph_tree(self,node,indent=''):
        """
        Recursively prints the tree structure of a graph starting from the specified node.

        Args:
        node (graphNode): The starting node from which to print the tree structure.
        indent (str): The indentation string to use for each level of the tree.
        """

        print(indent + '+-'+ str(node.unique) )
        child_indent = indent +'|' +' '*4



        if(len(node.out_nodes)==0):
            return
        else:
            for child in node.out_nodes:
                self.print_graph_tree(node=child, indent=child_indent)

    def save_image(self):
        _all_changes=self._traverse_graph()
        image_=self.image.copy().astype(np.int16)

        for key in _all_changes.keys():
            i,j,k=key
            image_[i][j][k]+=_all_changes[key] 
        image_=image_.astype(np.uint8)
        return image_


    def show_image(self):
        _all_changes=self._traverse_graph()
        image_=self.image.copy().astype(np.int16)

        for key in _all_changes.keys():
            i,j,k=key
            image_[i][j][k]+=_all_changes[key] 
        image_=image_.astype(np.uint8)

        plt.imshow(image_)
        plt.show()




import pickle
from datetime import datetime
import matplotlib.pyplot as plt
import os

def  save_graph(object_instance):
    graphs_dir = os.path.dirname(__file__)
    parent_dir = os.path.join(os.path.dirname(graphs_dir),'Graphs')

    name=object_instance.graph_name
    path=parent_dir+"/"+name+".pkl"

    with open(path, 'wb') as output:
       pickle.dump(object_instance,output,pickle.HIGHEST_PROTOCOL)


def load_graph(graph_name):

    '''
    function to load imageGraph object from pickle file

    input-> graph_name , graph_name attribute of imageGraph instance
    output-> imageGraph object
    '''
    graphs_dir = os.path.dirname(__file__)
    parent_dir = os.path.join(os.path.dirname(graphs_dir),'Graphs')

    path=parent_dir+'/'+graph_name+".pkl"
    
    with open(path,'rb') as input:
        graph=pickle.load(input)
    return graph


def show_image(obj):
   image_array= obj.image

   plt.imshow(image_array)
   plt.title("Displayed Image")
   plt.show()






import requests
import os

def upload_file(filename, server_address):
    url = f'http://{server_address}'
    headers = {'File-Name': os.path.basename(filename)}
    with open(filename, 'rb') as f:
        response = requests.post(url, headers=headers, data=f)
    print(response.text)

if __name__ == "__main__":
    filename_= input("Enter the name of the file you want to push: ").strip('"')
    path="C:\\Users\\Lenovo India\\OneDrive\\Desktop\\PixV repo\\sudo-rm-rf\\src\\Graphs\\"
    filename=path+filename_
    # filename = input("Enter the path of the file you want to push: ").strip('"')

    # server_address = input("Enter the server address (e.g., 127.0.0.1:8000): ")
    server_address="127.0.0.1:8000"
    upload_file(filename, server_address)


import requests
import os

def download_file(filename, server_address, download_dir):
    url = f'http://{server_address}/{filename}'
    response = requests.get(url)
    if response.status_code == 200:
        download_path = os.path.join(download_dir, filename)
        with open(download_path, 'wb') as f:
            f.write(response.content)
        print(f"File '{filename}' pulled successfully to '{download_path}'")
    else:
        print("File not found on the server")

# if __name__ == "__main__":
#     path="C:\\Users\\Lenovo India\\OneDrive\\Desktop\\PixV repo\\sudo-rm-rf\\src\\Graphs\\"
#     filename=input("Enter the name of the file you want to pull :")
#     # filename_=path+filename
#     # filename = input("Enter the name of the file you want to pull: ")

#     # server_address = input("Enter the server address (e.g., 127.0.0.1:8000): ")
#     server_address="127.0.0.1:8000"
#     # download_dir = input("Enter the path where you want to download the file: ")
#     # download_dir="C:\Users\Lenovo India\OneDrive\Desktop\PixV repo\sudo-rm-rf\src\Graphs"
#     download_file(filename, server_address, path)